#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

/* local includes */
#include "iso.h"
#include "util.h"

/* local definitions */
/* #define PACKAGE */
/* #define VERSION */
#define READ_COMMAND	(1)
#define COMPARE_COMMAND (2)
#define TEST_COMMAND	(3)
#define INFO_COMMAND	(4)
#define SIZE_COMMAND	(5)

/* global options */
static char *arg_device;
static int arg_blocksize;
static int arg_ignerr;
static int arg_command;
static int arg_verbosity;

/* other global variables */
static int displayed_status;
static FILE *message_out;
static char *compare_mem;
static int errors;
static char *prgname;

/* prototypes */
void interrupt_signal (int);
void clean_up_read_status (void);
void display_read_status (int, int);
void error_exit (char *);
void write_block (int, char *, int, int, char *);
void read_block (int, char *, int, int, char *);
void compare_block (int, char *, int, int, char *);

void
display_read_status (bytes_read, bytes_total)
  int bytes_read;
  int bytes_total;
{
	static struct timeval last_timeval;
	static int last_bytes_read = -1;
	static int transfer_rate = -1;
	int blocks_total, blocks_read;
	char *cmd;
	
	if (!arg_verbosity)
		return;
	
	blocks_total = (bytes_total / arg_blocksize) + !!(bytes_total % arg_blocksize);
	blocks_read = (bytes_read / arg_blocksize) + !!(bytes_read % arg_blocksize);

	if (last_bytes_read != -1) {
		struct timeval tmp_timeval;

		if (gettimeofday(&tmp_timeval, NULL))
			error_exit(NULL);

		timeval_subtract2(&last_timeval, &tmp_timeval);

		if (!(tmp_timeval.tv_sec < 1 && tmp_timeval.tv_usec < 500000)) {
			/* update transfer_rate */
			transfer_rate = (bytes_read-last_bytes_read) / ((tmp_timeval.tv_sec*1000000 + tmp_timeval.tv_usec)/1000);
			last_bytes_read = -1;
		}
	}

	/* update (or initialize) structs for last remember */
	if (last_bytes_read == -1) {
		if (gettimeofday(&last_timeval, NULL))
			error_exit(NULL);
		last_bytes_read = bytes_read;
	}

	cmd = (arg_command == READ_COMMAND ? "Reading" : "Comparing");
	bytes_read /= 1024*1024;
	if (bytes_total) {
		bytes_total /= 1024*1024;
		if (transfer_rate != -1)
			fprintf(message_out,"%s %3d/%3d MB (%d/%d blocks) at %d kb/s    \r",
				cmd, bytes_read, bytes_total, blocks_read, blocks_total, transfer_rate);
		else
			fprintf(message_out,"%s %3d/%3d MB (%d/%d blocks)\r",
				cmd, bytes_read, bytes_total, blocks_read, blocks_total);
	} else {
		fprintf(message_out,"%s %3d     MB (%d blocks)\r", cmd, bytes_read, blocks_read);
	}

	displayed_status = 1;
}

void
error_exit (errmsg)
  char *errmsg;
{
	clean_up_read_status();

	if (errmsg)
		fprintf(stderr, "%s: %s\n", prgname, errmsg);
	else
		perror(prgname);

	exit(1);
}

void
clean_up_read_status (void)
{
	if (displayed_status) {
		displayed_status = 0;
		fprintf(message_out, "\n");
	}
}

void
interrupt_signal (sig)
  int sig;
{
	error_exit("interrupted");
}

int
main (argc, argv)
  int argc;
  char ** argv;
{
	/* local variables */
	int data_in, data_out, compare_in;
	int c_total, c_read, c_remain;
	int blocks;
	char *block_mem;
	struct timeval begin_time;
	char *write_file, *compare_file;

	/* initialize global variables depending on basename */
	/* remove path from argv[0] */
	prgname = strrchr(argv[0], '/');
	if (prgname)
		prgname = &prgname[1];
	else
		prgname = argv[0];

	if (!strcmp(prgname, "isosize"))
		arg_command = SIZE_COMMAND;
	else if (!strcmp(prgname, "isoinfo"))
		arg_command = INFO_COMMAND;
	else if (!strcmp(prgname, "isocmp") || !strcmp(prgname, "cmpiso"))
		arg_command = COMPARE_COMMAND;
	else {
		prgname = PACKAGE;
		arg_command = READ_COMMAND;
	}
	displayed_status = FALSE;
	errors = FALSE;
	arg_blocksize = 32*1024;
	arg_device = "/dev/cdrom";
	arg_ignerr = FALSE;
	arg_verbosity = 1;

	/* take care of arguments */
	while (TRUE) {
		int c;
		int option_index = 0;
		static struct option long_options[] = {
			{ "read",			no_argument,		NULL, 'r' },	/* def: TRUE */
			{ "compare",		no_argument,		NULL, 'c' },
			{ "test",			no_argument,		NULL, 't' },
			{ "info",			no_argument,		NULL, 'i' },
			{ "size",			no_argument,		NULL, 's' },
			{ "device", 		required_argument,	NULL, 'd' },	/* def: "/dev/cdrom" */
			{ "block-size", 	required_argument,	NULL, 'b' },	/* def: 32k */
		/*	{ "to-stdout", 		no_argument,		NULL, 'O' },	*/
			{ "ignore-errors",	no_argument,		NULL, 'e' },
		/*	{ "verbose",		no_argument,		NULL, 'v' },	*/
			{ "quiet",			no_argument,		NULL, 'q' },
			{ "silent", 		no_argument,        NULL, 'q' },    /* synonym for quiet */
			{ "version",		no_argument,		NULL, 'V' },	/* no one-char */
			{ "help",			no_argument,		NULL, 'H' },	/* no one-char */
			{ 0,				0,					0,	  0   }
		};

		c = getopt_long (argc, argv, "rctiseqd:b:", long_options, &option_index);
		if (c == EOF)
			break;

		switch (c) {
			case 'r': arg_command = READ_COMMAND;			break;
			case 'c': arg_command = COMPARE_COMMAND;		break;
			case 't': arg_command = TEST_COMMAND;			break;
			case 'i': arg_command = INFO_COMMAND;			break;
			case 's': arg_command = SIZE_COMMAND;			break;
		/*	case 'O': data_out = STDOUT_FILENO; 			break; */
			case 'f': arg_ignerr = TRUE; 					break;
		/*	case 'v': arg_verbosity = 2;					break; */
			case 'q': arg_verbosity = 0;					break;
			case 'd': arg_device = optarg; 					break;
			case 'b': arg_blocksize = parse_size(optarg);	break;
			case 'V':
				printf("%s %s\n", PACKAGE, VERSION);
				exit(0);
			case 'H':
				printf("Usage: %s [OPTION]... [FILE]\n", prgname);
				puts("Reads and compares images of ISO9660 file systems.");
				puts("\nMain operation mode:");
				puts("  -r, --read              read ISO FS and write to file (default)");
				puts("  -c, --compare           compare ISO FS with file");
				puts("  -t, --test              read ISO FS but do not write");
				puts("  -i, --info              output information on ISO FS and exit");
				puts("  -s, --size              output size of ISO FS and exit");
				puts("\nOperation modifiers:");
				puts("  -d, --device=FILE       device or file to read data from (/dev/cdrom)");
				puts("  -b, --block-size=SIZE   amount of data to read at a time (32k)");
			/*	puts("  -O, --to-stdout         write data to standard output"); */
				puts("  -f, --force             do not abort on read/write errors");
				puts("\nInformative output:");
			/*	puts("  -v, --verbose           explain more precise what is being done"); */
				puts("  -q, --quiet             display no progress information");
				puts("      --help              display this help and exit");
				puts("      --version           output version information and exit");
				puts("\nDefault device is `/dev/cdrom' and default block size is 32k. Last");
				puts("character of block size may be k or M to represent kilo- or megabytes");
				puts("(a kilobyte is defined as 1024 bytes, and a megabyte 1024k).");
				puts("If FILE argument is omitted, output to standard output");
				puts("\nReport bugs to <osk@hem.passagen.se>");
				exit(0);
			case '?':
			default:
				exit(1);
		}
	}

	/* set up signals */
	{
		struct sigaction action;

		memset(&action, 0, sizeof(struct sigaction));
		action.sa_handler = interrupt_signal;
		sigemptyset(&action.sa_mask);
		sigaddset(&action.sa_mask, SIGINT);
		sigaction(SIGINT, &action, NULL);
	}

	/* initialize files and open devices */
	if ((data_in = open(arg_device, 0)) == -1) {
		clean_up_read_status();
		fprintf(stderr, "%s: %s: %s\n", prgname, arg_device, strerror(errno));
		exit(1);
	}
	/* (apparently stdout has different buffering than stderr, so use stderr always) */
	message_out = stderr;
	if (arg_command == COMPARE_COMMAND) {
		if (optind >= argc) {
			compare_in = STDIN_FILENO;
			compare_file = NULL;
			/*error_exit("missing file argument");*/
		} else {
			compare_file = argv[optind];
			if ((compare_in = open(compare_file, O_RDONLY, 0666)) == -1) {
				clean_up_read_status();
				fprintf(stderr, "%s: %s: %s\n", prgname, argv[optind], strerror(errno));
				exit(1);
			}
		}
	} else if (arg_command == READ_COMMAND) {
		if (optind >= argc) {
			/*message_out = stderr;*/
			data_out = STDOUT_FILENO;
			write_file = NULL;
		} else {
			write_file = argv[optind];
			if ((data_out = open(write_file, O_WRONLY|O_CREAT|O_TRUNC, 0666)) == -1) {
				clean_up_read_status();
				fprintf(stderr, "%s: %s: %s\n", prgname, argv[optind], strerror(errno));
				exit(1);
			}
		}
	}

	/* how many full blocks do we need to read to cover the iso header? */
	blocks =
		((ISO_HEADER_OFFSET+sizeof(struct iso_primary_descriptor)) / arg_blocksize)
		+ !!((ISO_HEADER_OFFSET+sizeof(struct iso_primary_descriptor)) % arg_blocksize);

	/* allocate that much memory + comparision memory */
	c_remain = blocks * arg_blocksize;
	if (!(block_mem = (char *) malloc(c_remain
	  + (arg_command == COMPARE_COMMAND ? arg_blocksize:0)))) {
		errno = ENOMEM;
		error_exit(NULL);
	}
	if (arg_command == COMPARE_COMMAND)
		compare_mem = &block_mem[c_remain];

	/* initialize timer */
	if (arg_command != SIZE_COMMAND && arg_command != INFO_COMMAND) {
		if (gettimeofday(&begin_time, NULL)) {
			error_exit(NULL);
		}
	}

	/* read first blocks */
	c_read = 0;
	while (c_remain >= arg_blocksize) {
		if (arg_command != SIZE_COMMAND && arg_command != INFO_COMMAND)
			display_read_status(c_read + arg_blocksize, 0);
		read_block(data_in, &block_mem[c_read], c_read, arg_blocksize, arg_device);
		c_read += arg_blocksize;
		c_remain -= arg_blocksize;
	}

	/* get the iso size from read block */
	{
		struct iso_primary_descriptor *ipd;
		ipd = (struct iso_primary_descriptor *) (block_mem+ISO_HEADER_OFFSET);
		c_remain = c_total = isonum_733(ipd->volume_space_size) * 2048;
	}
	if (arg_command == SIZE_COMMAND) {
		printf("%d\n", c_total);
		exit(0);
	} else if (arg_command == INFO_COMMAND) {
		printf("size=%d\n", c_total);
		exit(0);
	}

	/* write first block(s) */
	if (c_remain <  blocks*arg_blocksize) {
		if (arg_command == READ_COMMAND)
			write_block(data_out, block_mem, 0, c_remain, write_file);
		else if (arg_command == COMPARE_COMMAND)
			compare_block(compare_in, block_mem, 0, c_remain, compare_file);
		c_remain = 0;
	} else {
		while (blocks) {
			if (arg_command == READ_COMMAND)
				write_block(data_out, &block_mem[c_total-c_remain], c_total - c_remain, arg_blocksize, write_file);
			else if (arg_command == COMPARE_COMMAND)
				compare_block(compare_in, &block_mem[c_total-c_remain], c_total - c_remain, arg_blocksize, compare_file);
			blocks--;
			c_remain -= arg_blocksize;
		}
	}

	/* main loop - repeat until all blocks copied */
	while (c_remain >= arg_blocksize) {
		/* read block */
		display_read_status(c_total - c_remain + arg_blocksize, c_total);
		read_block(data_in, block_mem, c_read, arg_blocksize, arg_device);

		/* write block */
		if (arg_command == READ_COMMAND)
			write_block(data_out, block_mem, c_read, arg_blocksize, write_file);
		else if (arg_command == COMPARE_COMMAND)
			compare_block(compare_in, block_mem, c_read, arg_blocksize, compare_file);
		c_read += arg_blocksize;
		c_remain -= arg_blocksize;
	}

	/* read and write remaining data if any */
	if (c_remain > 0) {
		/* read remaining */
		display_read_status(c_total, c_total);
		read_block(data_in, block_mem, c_read, c_remain, arg_device);

		/* write remaining */
		if (arg_command == READ_COMMAND)
			write_block(data_out, block_mem, c_read, c_remain, write_file);
		else if (arg_command == COMPARE_COMMAND)
			compare_block(compare_in, block_mem, c_read, c_remain, compare_file);
	}

	free(block_mem);
	if (arg_command == READ_COMMAND)
		close(data_out);
	else if (arg_command == COMPARE_COMMAND)
		close(compare_in);
	close(data_in);

	/* display ending status */
	{
		char *cmd;
		struct timeval tmp_timeval;

		if (gettimeofday(&tmp_timeval, NULL))
			error_exit(NULL);
		timeval_subtract2(&begin_time, &tmp_timeval);
		clean_up_read_status();
		if (arg_verbosity)
			if (!errors) {
				cmd = (arg_command == READ_COMMAND ? "read" : "compared");
				fprintf(message_out, "Successfully %s %d MB (%d blocks) at an average of %d kb/s\n",
					cmd, c_total/(1024*1024),
					(c_total / arg_blocksize) + !!(c_total % arg_blocksize),
					c_total / ((tmp_timeval.tv_sec*1000000 + tmp_timeval.tv_usec)/1000));
			} else {
				cmd = (arg_command == READ_COMMAND ? "Read" : "Compared");
				fprintf(message_out, "%s %d MB (%d blocks) at an average of %d kb/s. Errors occured during process.\n",
					cmd, c_total/(1024*1024),
					(c_total / arg_blocksize) + !!(c_total % arg_blocksize),
					c_total / ((tmp_timeval.tv_sec*1000000 + tmp_timeval.tv_usec)/1000));
			}
	}

	exit(!errors);
}

void
write_block (fd, mem, offset, length, device)
  int fd;
  char *mem;
  int offset;
  int length;
  char *device;
{
	if (xwrite(fd, mem, length) == -1) {
		clean_up_read_status();
		fprintf(stderr, "%s: %s%s write error at offset %d: %s\n",
			prgname, (device?device:""),(device?": ":""), offset, strerror(errno));
		if (arg_ignerr) {
			errors = TRUE;
			if (lseek(fd, offset+length, SEEK_SET) != offset+length)
				error_exit(NULL);
		} else {
			exit(1);
		}
	}
}

void
compare_block (fd, mem, offset, length, device)
  int fd;
  char *mem;
  int offset;
  int length;
  char *device;
{
	/* first read the block */
	read_block(fd, compare_mem, offset, length, device);

	/* compare the memory */
	if (memcmp(mem, compare_mem, length)) {
		clean_up_read_status();
		fprintf(stderr, "%s: data differ at offset %d: %s\n",
			prgname, offset, strerror(errno));
		exit(1);
	}
}

void
read_block (fd, mem, offset, length, device)
  int fd;
  char *mem;
  int offset;
  int length;
  char *device;
{
	int result;

	result = xread(fd, mem, length);
	if (result <= 0) {
		clean_up_read_status();
		fprintf(stderr, "%s: %s%sread error at offset %d: %s\n",
			prgname, (device?device:""),(device?": ":""), offset, result ? strerror(errno) : "premature end");
		if (arg_ignerr) {
			errors = TRUE;
			if (lseek(fd, offset+length, SEEK_SET) != offset+length)
				error_exit(NULL);
		} else {
			exit(1);
		}
	}
}
